/**
 * FileParserHandler.cpp
 * Author: Yuan Wang
 * Document and error handler for xml parser.
 */

#include <string.h>
#include <iostream.h>
#include <stdlib.h>

#include <util/XMLUni.hpp>
#include <sax/AttributeList.hpp>

#include "FileParserHandler.h"


/**
 * Constructor.
 */
FileParserHandler::FileParserHandler ()
{
  this->fid = -1;
}


/**
 * Destructor.
 */
FileParserHandler::~FileParserHandler ()
{
}


/**
 * Set Ying Wu and Wei Shen's Object
 */
void FileParserHandler::setStorageObject ( FileStore* fileStore,
					   RecordStore* recordStore,
					   Sort* sort )
{
  this->fileStore = fileStore;
  this->recordStore = recordStore;
  this->sort = sort;
}


/**
 * Set fid.
 */
void FileParserHandler::setFileId ( int fid )
{
  this->fid = fid;
  //cout << "To parse XML File:\t" << this->fid << endl;
}


/**
 * Set names of elements that are being indexed. and open buffer
 */
void FileParserHandler::setIndexName ( char* rootname, char* indexnames[],
				       int indexcount, int* fieldLength )
{
  this->rootName = rootname;
  //  this->indexNames = indexnames;
  this->indexCount = indexcount;

  //cout << "Root:\t" << this->rootName << endl;
  for ( int i = 0; i < this->indexCount; i++ )
    {
      this->indexNames[i] = indexnames[i];
      //cout << "Index#" << i << ":\t" << this->indexNames[i] << endl;
    }

  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  this->indexValue = new char**[indexcount];
  for( int i = 0; i < indexcount; i++ )
    this->indexValue[i] = new char*[MAX_INDEX_VALUE_COUNT];
  this->indexValueCount = new int[indexcount];
  */

  this->indexId = -1;
}


/**
 * Error handler.
 * @param e  exception
 */
void FileParserHandler::error ( const SAXParseException& e )
{
  cerr << "\nError at (file " << StrX( e.getSystemId() ) << ", line "
       << e.getLineNumber() << ", char " << e.getColumnNumber()
       << "): " << StrX( e.getMessage() ) << endl;
}


/**
 * Fatal error handler.
 * @param e  exception
 */
void FileParserHandler::fatalError ( const SAXParseException& e )
{
  cerr << "\nFatal Error at (file " << StrX( e.getSystemId() ) << ", line "
       << e.getLineNumber() << ", char " << e.getColumnNumber()
       << "): " << StrX( e.getMessage() ) << endl;
}


/**
 * Warning handler.
 * @param e  exception
 */
void FileParserHandler::warning ( const SAXParseException& e )
{
  cerr << "\nWarning at (file " << StrX( e.getSystemId() ) << ", line "
       << e.getLineNumber() << ", char " << e.getColumnNumber()
       << "): " << StrX( e.getMessage() ) << endl;
}


/**
 * DTD handler.
 */
void FileParserHandler::unparsedEntityDecl ( const XMLCh* const name,
					     const XMLCh* const publicId,
					     const XMLCh* const systemId,
					     const XMLCh* const notationName )
{
  // Not used at this time.
}


/**
 * Another DTD handler.
 */
void FileParserHandler::notationDecl ( const XMLCh* const name,
					const XMLCh* const publicId,
					const XMLCh* const systemId )
{
  // Not used at this time.
}


/**
 * Document handler.
 */
void FileParserHandler::characters ( const XMLCh* const chars,
				     const unsigned int length )
{
  if ( this->inRecord )
    {
      StrX* strx = new StrX( chars, length );

      this->elementContent = strdup( strx->localForm() );

      delete strx;
    }
}


/**
 * Ignore white space.
 */
void FileParserHandler::ignorableWhitespace ( const XMLCh* const chars,
					      const unsigned int length )
{
  // Do nothing. i.e., Do not ignore white space.
}


/**
 * Receive processing instructions.
 */
void FileParserHandler::processingInstruction ( const XMLCh* const target,
						const XMLCh* const data )
{
  // Do nothing.
}


/**
 * Start a document.
 */
void FileParserHandler::startDocument ()
{
  this->recordCount = 0;
  this->inRecord = false;
  this->sameLevel = false;

  this->fid++;
}


/**
 * Initialize before reading a record.
 */
void FileParserHandler::initRecord ()
{
  this->doIndex = false;

  for( int i = 0; i < this->indexCount; i++ )
    {
      for( int j = 0; j < this->indexValueCount[i]; j++ )
	free( this->indexValue[i][j] );
      this->indexValueCount[i] = 0;
    }
}


/**
 * End a document.
 */
void FileParserHandler::endDocument ()
{
  // Do nothing.
  //cout << "End of document!" << endl;
}


/**
 * Start an element.
 */
void FileParserHandler::startElement ( const XMLCh* const name,
				       AttributeList& attributes )
{
  StrX* strx = new StrX( name );
  char* realName = strdup( strx->localForm() );
  // cout << "Start element:\t" << realName << endl;

  // Free last element.
  //  if ( this->elementContent != NULL )
  //    free( this->elementContent );
  
  // Find start of a record.
  if ( !this->inRecord )
    {
      if ( strcmp( realName, this->rootName ) == 0 )
	{
	  // Notify YingWu.
	  //cout << "Create record" << endl;
	  this->inRecord = true;
	  this->recordStore->createRecord( realName, this->fid, this->recordCount );
	}
    }
  else
    {
      // Create element in Ying Wu's storage.
      if ( !this->sameLevel )
	{
	  //cout << "Go to next level" << endl;
	  this->recordStore->goToNextLevel();
	}
      else
	this->sameLevel = false;

      //cout << "Create element\t" << realName << endl;
      this->recordStore->createElement( realName );

      this->doIndex = false;

      // Set index on elements.
      for ( int i = 0; i < this->indexCount; i++ )
	{
	  // Should use a hash table here.
	  if ( strcmp( realName, this->indexNames[i] ) == 0 )
	    {
	      this->doIndex = true;
	      this->indexId = i;
	      break;
	    }
	}
    }

  free( realName );
  delete strx;

  // Set index on attributes.
  unsigned int len = attributes.getLength();

  for ( unsigned int index = 0; index < len; index++ )
    {
      StrX* strx1 = new StrX( attributes.getName( index ) );
      char* attrName = strdup( strx1->localForm() );
      StrX* strx2 = new StrX( attributes.getValue( index ) );
      char* attrValue = strdup( strx2->localForm() );

      //cout << "Add attribute:\t" << attrName << "\t" << attrValue << endl;

      // attribute is counted from 1, sigh...
      this->recordStore->addAttribute( index+1, attrName, attrValue );

      free( attrName );
      free( attrValue );
      delete strx1;
      delete strx2;
    }
}


/**
 * End an element.
 */
void FileParserHandler::endElement ( const XMLCh* const name )
{
  //  this->doIndex = false;

  //cout << "End of element" << endl;

  if ( this->inRecord )
    {
      StrX* strx = new StrX( name );
      char* realName = strdup( strx->localForm() );

      // The end of current record.
      if ( strcmp( realName, this->rootName ) == 0 ){
	  if ( this->sameLevel )
	    {
	      // Finish this level.
	      //cout << "Go back to previous level" << endl;
	      this->recordStore->finishThisLevel();
	    }

	  // Finish record and inserted into the file.
	  int recordSize = 0;
	  int wfid;
	  unsigned short wrid;

	  //cout << "Finish record" << endl;

	  char* record = this->recordStore->finishRecord( recordSize );

	  //cout << "Insert record" << endl;

	  //if ( this->fileStore == NULL )
	  //  exit(0);
	  
	  if ( !this->fileStore->insertRecord( record, recordSize, wfid, wrid ) )
	    {
	      //cout << "Error: inserting record to file!" << endl;
	      exit( -1 );
	    }

	  delete [] record;

	  // Pass indices to Weishen!!

	  for( int i = 0; i < this->indexCount; i++ )
	    {
	      for( int j = 0; j < this->indexValueCount[i]; j++ )
		{
		  this->sort->insertRecord( i, this->fid, this->recordCount, wfid, wrid, this->indexValue[i][j] );
/*

			if(i==0) {
				cout<<"---"<<this->indexValue[i][j]<<endl;
			}
			*/
		  //		  //cout << "Insert index:\t" << i << "\t" << this->fid << "\t"
		  //		       << this->recordCount << "\t" << wfid << "\t" << wrid << "\t"
		  //		       << this->indexValue[i][j] << endl;
		  free( this->indexValue[i][j] );
		}

	      this->indexValueCount[i] = 0;
	    }

	  free( realName );
	  delete( strx );

	  this->recordCount++;
	  this->inRecord = false;
	  this->sameLevel = false;

	  free( this->elementContent );
	  return;
	}

      if ( this->sameLevel == false )
	{
	  //cout << "Add element content\t" << this->elementContent << endl;
	  this->recordStore->addElementContent( this->elementContent );

	  // Make index
	  if ( this->doIndex )
	    {
	      this->indexValue[this->indexId][this->indexValueCount[this->indexId]]
		= this->elementContent;
	      this->indexValueCount[this->indexId]++;
	      this->elementContent = NULL;
	      this->doIndex = false;
	    }
	  else
	    free( this->elementContent );
	}
      else
	{
	  free( this->elementContent );
	}

      // Finish this element.
      //cout << "Finish Element\t" << endl;
      this->recordStore->finishElement();

      if ( this->sameLevel )
	{
	  // Finish this level.
	  //cout << "Go back to previous level" << endl;
	  this->recordStore->finishThisLevel();
	}
      else
	{
	  this->sameLevel = true;
	}
    }
}



/**
 * StrX: Private helper methods
 */
void StrX::transcode( const XMLCh* const toTranscode, const unsigned int len )
{
  // Short circuit if its a null pointer
  if ( !toTranscode || ( !toTranscode[0] ) )
    {
      fLocalForm = new char[1];
      fLocalForm[0] = 0;
      return;
    }

  // See if our XMLCh and wchar_t as the same on this platform
  const bool isSameSize = ( sizeof( XMLCh ) == sizeof( wchar_t ) );

  //
  //  Get the actual number of chars. If the passed len is zero, its null
  //  terminated. Else we have to use the len.
  //
  wchar_t realLen = ( wchar_t )len;
  if ( !realLen )
    {
      //
      //  We cannot just assume we can use wcslen() because we don't know
      //  if our XMLCh is the same as wchar_t on this platform.
      //
      const XMLCh* tmpPtr = toTranscode;
      while ( *( tmpPtr++ ) )
	realLen++;
    }

  //
  //  If either the passed length was non-zero or our char sizes are not 
  //  same, we have to use a temp buffer. Since this is common in these
  //  samples, we just do it anyway.
  //
  wchar_t* tmpSource = new wchar_t[realLen + 1];
  if ( isSameSize )
    {
      memcpy( tmpSource, toTranscode, realLen * sizeof( wchar_t ) );
    }
  else
    {
      for ( unsigned int index = 0; index < realLen; index++ )
	tmpSource[index] = ( wchar_t )toTranscode[index];
    }
  tmpSource[realLen] = 0;

  // See now many chars we need to transcode this guy
  const unsigned int targetLen = ::wcstombs( 0, tmpSource, 0 );

  // Allocate out storage member
  fLocalForm = new char[targetLen + 1];

  //
  //  And transcode our temp source buffer to the local buffer. Cap it
  //  off since the converter won't do it (because the null is beyond
  //  where the target will fill up.)
  //
  ::wcstombs( fLocalForm, tmpSource, targetLen );
  fLocalForm[targetLen] = 0;

  // Don't forget to delete our temp buffer
  delete [] tmpSource;
}
